-- Arithmetic Unit Module

library ieee;

use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

 
package arith_pckg is

     component arithmetic_unit is

	port(

		y0, y1, y2, y3: in bit;

		As: in std_logic_vector(3 downto 0); -- input RR4 sign bit register    |  1st
		A1: in std_logic_vector(3 downto 0); -- input RR4 higher bit register  |  input
		A0: in std_logic_vector(3 downto 0); -- input RR4 lower bit register   |  operand

		Bs: in std_logic_vector(3 downto 0); -- input RR4 sign bit register    |  2nd
		B1: in std_logic_vector(3 downto 0); -- input RR4 higher bit register  |  input
		B0: in std_logic_vector(3 downto 0); -- input RR4 lower bit register   |  operand

		Ts: out std_logic_vector(4 downto 0); -- output RR4 addition register (sign bit)
		T1: out std_logic_vector(4 downto 0); -- output RR4 addition register (higher bit)
		T0: out std_logic_vector(4 downto 0); -- output RR4 addition register (lower bit)

		Ms: out std_logic_vector(7 downto 0); -- output RR4 multiplication register (sign bit)
		M1: out std_logic_vector(7 downto 0); -- output RR4 multiplication register (higher bit)
		M0: out std_logic_vector(7 downto 0)  -- output RR4 multiplication register (lower bit)
	);
     end component arithmetic_unit;

     subtype vector is std_logic_vector(3 downto 0);

end package arith_pckg;


library ieee;

use ieee.std_logic_1164.all;
use ieee.numeric_std.all;


entity arithmetic_unit is

	port(

		y0, y1, y2, y3: in bit;

		As: in std_logic_vector(3 downto 0); -- input RR4 sign bit register    |  1st
		A1: in std_logic_vector(3 downto 0); -- input RR4 higher bit register  |  input
		A0: in std_logic_vector(3 downto 0); -- input RR4 lower bit register   |  operand

		Bs: in std_logic_vector(3 downto 0); -- input RR4 sign bit register    |  2nd
		B1: in std_logic_vector(3 downto 0); -- input RR4 higher bit register  |  input
		B0: in std_logic_vector(3 downto 0); -- input RR4 lower bit register   |  operand

		Ts: out std_logic_vector(4 downto 0); -- output RR4 addition register (sign bit)
		T1: out std_logic_vector(4 downto 0); -- output RR4 addition register (higher bit)
		T0: out std_logic_vector(4 downto 0); -- output RR4 addition register (lower bit)

		Ms: out std_logic_vector(7 downto 0); -- output RR4 multiplication register (sign bit)
		M1: out std_logic_vector(7 downto 0); -- output RR4 multiplication register (higher bit)
		M0: out std_logic_vector(7 downto 0)  -- output RR4 multiplication register (lower bit)
	);
	
end entity arithmetic_unit;


architecture behav of arithmetic_unit is

	
	component RR4_adder is

	port(
		Scp: in bit;
		Xs, X1, X0: in bit;
		Ys, Y1, Y0: in bit;
		
		Sc, C0, C1: out bit;  -- intermediate carry digit
		Ss, S0, S1: out bit   -- intermediate sum digit
		
	);
     end component RR4_adder;
     
     component partial_product is

	port(

		Sa, a0, a1: in bit;  -- RR4 intermediate digit
		Sb, b0, b1: in bit;  -- RR4 intermediate digit

		Sc, c0, c1: in bit   -- RR4 partial product

	);
     end component partial_product;

	signal Sic: std_logic_vector(3 downto 0); -- RR4 intermediate carry digits for addition operation (sign bit)
	signal ic1: std_logic_vector(3 downto 0); -- RR4 intermediate carry digits for addition operation (higher bit)
	signal ic0: std_logic_vector(3 downto 0); -- RR4 intermediate carry digits for addition operation (lower bit)

	signal Sis: std_logic_vector(3 downto 0); -- RR4 intermediate sum digits for addition operation (sign bit)
	signal is1: std_logic_vector(3 downto 0); -- RR4 intermediate sum digits for addition operation (higher bit)
	signal is0: std_logic_vector(3 downto 0); -- RR4 intermediate sum digits for addition operation (lower bit)

	signal Sc: vector(3 downto 0);  -- RR4 intermediate carry digits (sign bit)
	signal c1: vector(3 downto 0);  -- RR4 intermediate carry digits (higher bit)   
	signal c0: vector(3 downto 0);  -- RR4 intermediate carry digits (lower bit)	

	signal Scp: vector(3 downto 0);  -- RR4 intermediate previous carry digits (sign bit)
	signal c1p: vector(3 downto 0);  -- RR4 intermediate previous carry digits (higher bit)   
	signal c0p: vector(3 downto 0);  -- RR4 intermediate previous carry digits (lower bit)	

	signal Ss: vector(3 downto 0);  -- RR4 intermediate sum digits (sign bit)
	signal s1: vector(3 downto 0);  -- RR4 intermediate sum digits (higher bit)   
	signal s0: vector(3 downto 0);  -- RR4 intermediate sum digits (lower bit)
	
	signal Ps: vector(3 downto 0); -- RR4 partial products (sign bit)
	signal P1: vector(3 downto 0); -- RR4 partial products (higher bit)   
	signal P0: vector(3 downto 0); -- RR4 partial products (lower bit)


begin

	arith_op:process(Sic, ic1, ic0, Sis, is1, is0) is
	
	begin
	 
 	if y0 ='1' then     -- addition operation

		i1:for i in 0 to 3 generate

		i2:if i='0' generate

		ID_i: RR4_adder

			  port map (
				   Scp=>'0',
				   Xs=>As(i), X1=>A1(i), X0=>A0(i),
				   Ys=>Bs(i), Y1=>B1(i), Y0=>B0(i),
				   
			           Sc=>Sic(i), C0=>ic1(i), C1=>ic0(i),
				   Ss=>Sis(i), S0=>is1(i), S1=>is0(i)
				  );

		FS_i: partial_product

			  port map(
				   Sis(i),is1(i),is0(i),
				   '0','0','0',

				   Ts(i),T1(i),T0(i)
				  );

		else

		ID_i: RR4_adder(behav)

			  port map(
				   Sic(i-1),
				   As(i),A1(i),A0(i),
				   Bs(i),B1(i),B0(i),
				   
			           Sic(i),ic1(i),ic0(i),
				   Sis(i),is1(i),is0(i)
				  );

		FS_i: partial_product(behav)

			  port map(
				   Sis(i),is1(i),is0(i),
				   Sic(i-1),ic1(i-1),ic0(i-1),

				   Ts(i),T1(i),T0(i)
				  );			
		
		end generate i2;

		end generate i1;

		Ts(4)<=Sic(3);
		T1(4)<=ic1(3);
		T0(4)<=ic0(3);

	elsif y1 ='1' then    --subtraction operation

		i3:for i in 0 to 3 generate

		i4:if i='0' generate

		ID_i: RR4_adder

			  port map(
				   '0',
				   As(i),A1(i),A0(i),
				   (not Bs(i)),B1(i),B0(i),
				   
			           Sic(i),ic1(i),ic0(i),
				   Sis(i),is1(i),is0(i)
				  );

		FS_i: partial_product

			  port map(
				   Sis(i),is1(i),is0(i),
				   '0','0','0',

				   Ts(i),T1(i),T0(i)
				  );

		else

		ID_i: RR4_adder

			  port map(
				   Sic(i-1),
				   As(i),A1(i),A0(i),
				   (not Bs(i)),B1(i),B0(i),
				   
			           Sic(i),ic1(i),ic0(i),
				   Sis(i),is1(i),is0(i)
				  );

		FS_i: partial_product

			  port map(
				   Sis(i),is1(i),is0(i),
				   Sic(i-1),ic1(i-1),ic0(i-1),

				   Ts(i),T1(i),T0(i)
				  );			
		
		end generate i4;

		end generate i3;

		Ts(4)<=Sic(3);
		T1(4)<=ic1(3);
		T0(4)<=ic0(3);
		

	elsif y2 ='1' then      --multiplication operation

		i5:for i in 0 to 3 generate

			i6:for j in 0 to 3 generate

				i7:if j ='0' generate

				ID_i_j: RR4_multiplier  -- intermediate RR4 digits generation

				port map(
					 '0', 
					 As(j),A1(j),A0(j),
					 Bs(i),B1(i),B0(i),

					 Ss(i),s1(i),s0(i),
					 Sc(i),c1(i),c0(i)
					);
				
				PP_i_j: partial_product  -- RR4 partial products generation

				port map(
					 Ss(i),s1(i),s0(i),
					 '0','0','0',

					 Ps(i),P1(i),P0(i)
					);

				Scp(i)<=Sc(i);
				c1p(i)<=c1(i);
				c0p(i)<=c0(i);
				
				else

				ID_i_j: RR4_multiplier  -- intermediate RR4 digits generation

				port map(
					 Scp(i), 
					 As(j),A1(j),A0(j),
					 Bs(i),B1(i),B0(i),

					 Ss(i),s1(i),s0(i),
					 Sc(i),c1(i),c0(i)
					);


				PP_i_j: partial_product  -- RR4 partial products generation

				port map(
					 Ss(i),s1(i),s0(i),
					 Scp(i),c1p(i),c0p(i),

					 Ps(i),P1(i),P0(i)
					);

				end generate i7;
				
				Scp(i)<=Sc(i);
				c1p(i)<=c1(i);
				c0p(i)<=c0(i);
				
			end generate i6;
			
				Ps(i)<= Sc(i);
				P1(i)<= c1(i);
				P0(i)<= c0(i);

		end generate i5;



		
	elsif y3 ='1' then      --complementation operation

		i8:for i in 0 to 3 generate

		Ts(i)<= not As(i);
		T1(i)<= A1(i);
		T0(i)<= A0(i);

		end generate i8;

		Ts(4)<='0';
		T1(4)<='0';
		T0(4)<='0';

	end if;
	
	end process arith_op;

end architecture behav;		
